/*
 * $Id: $
 *
 * Copyright (c) 2006 Network Appliance, Inc.
 * All rights reserved.
 */
package netapp.manage;

/**
 * Handle encoding character arrays into a base-16 string, and
 * decoding them again.
 * 
 * This is the encoding scheme ZAPI uses for "encrypted" values.
 */
public class Base16 {

    final static String hexChars = "0123456789abcdef";
    
    /**
     * Converts a Unicode hex character ('0', '1', ...) to an 
     * integer hex value (0, 1, 2).
     * 
     * @throws NumberFormatException if the character is not a hex character
     */
    static int hexValue(char ch) {
        ch &= 0xFF;
        if (ch >= '0' && ch <= '9') {
            return (ch - '0');
        }
        if (ch >= 'a' && ch <= 'f') {
            return 10 + (ch - 'a');
        }
        if (ch >= 'A' && ch <= 'F') {
            return 10 + (ch - 'A');
        }
        throw new NumberFormatException(
                "The base-16 input character " + ch + " is not a hexidecimal character");
    }
    
    /**
     * Converts the lower byte of an integer value into a hex character.
     */
    static char hexChar(int value) {
        value &= 0xF;
        return hexChars.charAt(value);
    }
    
    /**
     * Encodes some bytes using base-16.
     */
    public static String encode(char[] bytes) {
        if (bytes == null) {
            return null;
        }
        StringBuffer sb = new StringBuffer();
        
        for (int i = 0; i < bytes.length; i++) {
            char ch = bytes[i];
            sb.append(hexChar(ch>>4));
            sb.append(hexChar(ch));
        }
        return sb.toString();
    }
    
    /**
     * Decodes a string of base-16 characters into its bytes.
     */
    public static char[] decode(String text) {
        if (text == null) {
            return null;
        }
        if (0 != text.length() % 2) {
            throw new IllegalArgumentException(
                    "The base-16 input text must contain an even number of characters; " +
                    text + " has " + text.length() + " characters.");
        }
        int length = text.length() / 2;
        
        char[] result = new char[length];
        if (0 == length) {
            return result;
        }
        
        for (int i = 0; i < length; i++) {
            int ch1 = hexValue(text.charAt(2*i));
            int ch2 = hexValue(text.charAt(2*i + 1));
            result[i] = (char) ((ch1 << 4) | ch2);
        }
        return result;
    }
}
