/*
 * $Id: //depot/prod/zephyr/lufthansaair/src/java/netapp/manage/NaProxyServer.java#1 $
 *
 * Copyright (c) 2004 Network Appliance, Inc.
 * All rights reserved.
 *
 * ONTAPI Server stuff
 */

package netapp.manage;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.List;
import java.util.Iterator;

/**
 *  A class used to connect to proxy servers and invoke APIs.
 */
public class NaProxyServer extends NaServer {

	private String _target;
	private String _username;

	/**
	 *  Create a proxy server object with the given host address.
	 *
	 *  @param	address		the address of the proxy server.
	 *  @param	target		the name of the target server
	 */
	public NaProxyServer(InetAddress address, String target)
	{
		super(address);
		setProxyTarget(target);
	}

	/**
	 *  Create a proxy server object with the given hostname.
	 *
	 *  @param	hostname	the hostname of the proxy server
	 *  @param	target		the name of the target server
	 *  @exception	UnknownHostException	if the hostname is not known.
	 */
	public NaProxyServer(String hostname, String target)
		throws UnknownHostException
	{
		this(InetAddress.getByName(hostname), target);
	}

	/**
	 *  Set the target for this proxy server. The name is the hostname
	 *  for the machine that the proxy should forward the request to. It
	 *  could be an unqualified hostname, an fully-qualified domain name
	 *  (FQDN) or an IP address. The proxy server may support additional
	 *  types of target names, such as vfiler names.
	 *
	 *  @param	target		the name of the target server
	 */
	public void setProxyTarget(String target)
	{
		if (null == target) {
			throw new IllegalArgumentException("target cannot be null");
		}
		_target = target;
	}

	/**
	 *  Set the username for this proxy server. The proxy server will
	 *  attempt to use that user's account to execute the API.
	 *  <code>null</code> is a valid value that means the server should
	 *  use the highest privileged account available. Note that the
	 *  server may have a security policy which restricts which values
	 *  are valid for this field.
	 *
	 *  @param	username	the username of the account to use, or
	 *						<code>null</code>
	 */
	public void setProxyUsername(String username)
	{
		_username = username;
	}

	/**
	 *  Request that the proxy server invokes a single API request on the
	 *  target host.
	 *
	 *  @param	xi		The API element.
	 *  @return	Output results of the API.
	 *  @exception	IOException	if there is a communication problem
	 *  @exception	NaAuthenticationException	if the login was not accepted
	 *  @exception	NaProxyFailedException	if the proxy server returned a
	 *										failure result.
	 *  @exception	NaAPIFailedException	if the target server returned a
	 *										failure result.
	 *  @exception	NaProtocolException	if there is a protocol problem
	 */
	public NaElement invokeElem(NaElement xi) throws IOException,
		NaAuthenticationException, NaProxyFailedException, NaAPIFailedException,
		NaProtocolException
	{
		/*
		 *  Construct the proxy envelope for the API request
		 */
		NaElement nxi = new NaElement("api-proxy");
		nxi.addNewChild("target", _target);
		NaElement request = new NaElement("request");
		nxi.addChildElem(request);
		request.addNewChild("name", xi.getName());

		/*
		 *  Take the arguments to the API request and copy them into the
		 *  proxy wrapper.
		 */
		List argsList = xi.getChildren();
		if (0 != argsList.size()) {
			NaElement argsElem = new NaElement("args");
			request.addChildElem(argsElem);
			for (Iterator i = argsList.iterator(); i.hasNext(); ) {
				NaElement arg = (NaElement) i.next();
				argsElem.addChildElem(arg);
			}
		}

		NaElement xo;

		try {
			/*
			 *  Invoke the proxy-api with the wrapped request
			 */
			xo = super.invokeElem(nxi);
		} catch (NaAPIFailedException e) {
			/*
			 *  Here we deliberately change the exception type so that
			 *  callers can distinguish between API failures on the
			 *  proxy server and API failures on the target server.
			 */
			throw new NaProxyFailedException(e.getErrno(), e.getReason());
		}

		/*
		 *  Extract the response from the results (required element)
		 */
		NaElement response = xo.getChildByName("response");
		if (null == response) {
			throw new NaProtocolException("Missing proxy reponse");
		}

		/*
		 *  If the API succeeded, return the results now
		 */
		if ("passed".equals(response.getChildContent("status"))) {
			NaElement results = response.getChildByName("results");
			if (null == results) {
				throw new NaProtocolException("Missing proxy results");
			}
			return results;
		}

		/*
		 *  Otherwise, something went wrong, and we need to build and
		 *  throw an NaAPIFailedException.
		 */
		int errno;
		String s = response.getChildContent("errno");
		if (null == s) {
			throw new NaProtocolException("Missing errno in proxy response");
		}
		try {
			errno = Integer.parseInt(s);
		} catch (NumberFormatException e) {
			throw new NaProtocolException("Malformed errno code: " + s, e);
		}
		String reason = response.getChildContent("reason");
		if (null == reason) {
			throw new NaProtocolException("Missing reason in proxy response");
		}
		throw new NaAPIFailedException(errno, reason);
	}
}
