/*
 * $Id: //depot/prod/zephyr/lufthansaair/src/java/netapp/manage/NaServer.java#2 $
 *
 * Copyright (c) 2010 NetApp, Inc.
 * All rights reserved.
 *
 * API Server stuff
 */

package netapp.manage;

import java.io.IOException;
import java.io.StringReader;
import java.io.UnsupportedEncodingException;
import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.xml.sax.SAXException;

import netapp.manage.http.HTTPRequest;
import netapp.manage.http.HTTPResponse;
import netapp.manage.http.HTTPClient;
import netapp.manage.http.HTTPException;
import netapp.manage.http.AuthInfo;

import java.lang.reflect.Method;
import java.util.prefs.Preferences;

/**
 * A class used to connect to servers and invoke APIs.
 */
public class NaServer {

    public final static int SERVER_TYPE_FILER = 1;
    public final static int SERVER_TYPE_NETCACHE = 2;
    public final static int SERVER_TYPE_AGENT = 3;
    public final static int SERVER_TYPE_DFM = 4;
    public final static int SERVER_TYPE_OCUM = 5;

    public final static int TRANSPORT_TYPE_HTTP = 1;
    public final static int TRANSPORT_TYPE_HTTPS = 2;

    private final static int TCP_PORT_FILER = 443;
    private final static int TCP_PORT_NETCACHE = 80;
    private final static int TCP_PORT_AGENT = 4093;
    private final static int TCP_PORT_DFM = 8488;
    private final static int TCP_PORT_OCUM = 443;


    /*
      * authentication types
      */
    public final static int STYLE_LOGIN_PASSWORD = 1;
    public final static int STYLE_HOSTSEQUIV = 2;
    public final static int STYLE_RPC = 3;
    public final static int STYLE_CERTIFICATE = 4;

    public final static int KEYSTORE_TYPE_JKS = 1;
    public final static int KEYSTORE_TYPE_PKCS12 = 2;

    /**
     * @deprecated As of SDK version 1.4
     */
    public final static int LOGIN_PASSWORD = STYLE_LOGIN_PASSWORD;
    /**
     * @deprecated As of SDK version 1.4
     */
    public final static int HOSTSEQUIV = STYLE_HOSTSEQUIV;

    /*
      * API URI's
      */
    private final static String FILER_REQUEST_URI
            = "/servlets/netapp.servlets.admin.XMLrequest_filer";

    private final static String NETCACHE_REQUEST_URI
            = "/servlets/netapp.servlets.admin.XMLrequest";

    private final static String AGENT_REQUEST_URI
            = "/apis/XMLrequest";

    private final static String DFM_REQUEST_URI
            = "/apis/XMLrequest";

    private final static String FILER_XML_REQUEST
            = "<?xml version='1.0' encoding='utf-8' ?>\r\n"
            + "<!DOCTYPE netapp SYSTEM 'file:/etc/netapp_filer.dtd'>\r\n";

    private final static String AGENT_XML_REQUEST
            = "<?xml version='1.0' encoding='utf-8' ?>\r\n"
            + "<!DOCTYPE netapp SYSTEM 'file:/etc/netapp_agent.dtd'>\r\n";

    private final static String DFM_XML_REQUEST
            = "<?xml version='1.0' encoding='utf-8' ?>\r\n"
            + "<!DOCTYPE netapp SYSTEM 'file:/etc/netapp_dfm.dtd'>\r\n";

    private final static String FILER_NAMESPACE
            = "http://www.netapp.com/filer/admin";

    private final static String NMSDK_VERSION = "9.3";
    private final static String NMSDK_LANGUAGE = "Java";
    private static String NMSDK_PLATFORM = getPlatformInfo();
    private static String nmsdkAppName = null;

    private InetAddress _server;
    private String _serverName;
    private int _type;
    private int _transport;
    private String _login;
    private String _password;
    private int _style;
    private int _port;
    private int _major;
    private int _minor;
    private int _snoop;
    private int _debugStyle;
    private boolean _keepAlive;
    private List _sessions;
    private Thread _watcherThread;
    private String _vfiler;
    private String _rawXMLOutput = null;
    private int _timeOut;
	private String cookieValue;
    private AuthInfo _authInfo;
    private boolean _useTLS = true;
    private boolean _useSSL = false;
    private static Hashtable _tlsInUse  = new Hashtable();
    private boolean multipleApis = false;
    private boolean apiAttributeSet = false;

    /**
     * Create a server API object with server type SERVER_TYPE_FILER
     * and API version 1.0.
     *
     * @param    serverAddr        The IP address(IPv4 or IPv6) of the API server.
     */
    public NaServer(InetAddress serverAddr) {
        _server = serverAddr;
        _serverName = serverAddr.getHostName();
        _major = 1;
        _minor = 0;
        _login = null;
        _password = null;
        _type = SERVER_TYPE_FILER;
        _transport = TRANSPORT_TYPE_HTTPS; /* XXX HTTPS should be the default */
        _style = STYLE_LOGIN_PASSWORD;
        _port = 0;
        _snoop = 0;
        _keepAlive = false;
        _sessions = new ArrayList();
        _watcherThread = null;
        _vfiler = null;
        _debugStyle = 0;
        _rawXMLOutput = null;
        _timeOut = 0;
        _authInfo = new AuthInfo(_serverName);
    }

    /**
     * Create a server API object with server type SERVER_TYPE_FILER
     * and API version 1.0.
     *
     * @param    serverName        The name of the API server.
     * @exception UnknownHostException    if the hostname is not known.
     */
    /*
      * UnknownHostException is no longer effective.
      * It is a fix for burt 210951 and 216087.
      */
    public NaServer(String serverName)
            throws UnknownHostException {
        _serverName = serverName;
        _major = 1;
        _minor = 0;
        _login = null;
        _password = null;
        _type = SERVER_TYPE_FILER;
        _transport = TRANSPORT_TYPE_HTTP; /* XXX HTTPS should be the default */
        _style = STYLE_LOGIN_PASSWORD;
        _port = 0;
        _snoop = 0;
        _keepAlive = false;
        _sessions = new ArrayList();
        _watcherThread = null;
        _vfiler = null;
        _debugStyle = 0;
        _rawXMLOutput = null;
        _timeOut = 0;
        _authInfo = new AuthInfo(_serverName);
    }

    /**
     * Create a server API object with server type SERVER_TYPE_FILER
     *
     * @param    serverAddr        The IP address(IPv4 or IPv6) of the API server.
     * @param    majorVersion    The major API version number required, for
     * example "1" if version 1.0 is required.
     * @param    minorVersion    The minor API version number required, for
     * example "0" if version 1.0 is required.
     * <P> For DataFabric Manager APIs, pass 1.0 as the major
     * and minor version numbers.
     */
    public NaServer(InetAddress serverAddr, int majorVersion, int minorVersion) {
        this(serverAddr);
        setApiVersion(majorVersion, minorVersion);
    }

    /**
     * Create a server API object with server type SERVER_TYPE_FILER
     *
     * @param    serverName        The name of the API server.
     * @param    majorVersion    The major API version number required, for
     * example "1" if version 1.0 is required.
     * @param    minorVersion    The minor API version number required, for
     * example "0" if version 1.0 is required.
     * <P> For DataFabric Manager APIs, pass 1.0 as the major
     * and minor version numbers.
     * @exception UnknownHostException    if the hostname is not known.
     */
    /*
      * UnknownHostException is no longer effective.
      * It is a fix for burt 210951 and 216087.
      */
    public NaServer(String serverName, int majorVersion, int minorVersion)
            throws UnknownHostException {
        _serverName = serverName;
        _login = null;
        _password = null;
        _type = SERVER_TYPE_FILER;
        _transport = TRANSPORT_TYPE_HTTPS; /* XXX HTTPS should be the default */
        _style = STYLE_LOGIN_PASSWORD;
        _port = 0;
        _snoop = 0;
        _keepAlive = false;
        _sessions = new ArrayList();
        _watcherThread = null;
        _vfiler = null;
        setApiVersion(majorVersion, minorVersion);
        _debugStyle = 0;
        _rawXMLOutput = null;
        _timeOut = 0;
        _authInfo = new AuthInfo(_serverName);
    }

    /**
     * Set the API version for requests.
     *
     * @param    majorVersion    The major API version number required, for
     * example "1" if version 1.0 is required.
     * @param    minorVersion    The minor API version number required, for
     * example "0" if version 1.0 is required.
     * <P> For DataFabric Manager APIs, pass 1.0 as the major
     * and minor version numbers.
     */
    public void setApiVersion(int majorVersion, int minorVersion) {
        _major = majorVersion;
        _minor = minorVersion;
    }

    /**
     * Close any connections used by this object.
     */
    public synchronized void close() {
        setKeepAliveEnabled(false);
        if (null != _watcherThread) {
            _watcherThread.interrupt();
        }
    }

    /**
     * Turn HTTP keep alives on or off. The default is off. If keep alives
     * are enabled, you must call close() when you are done using this object.
     *
     * @param    enabled        Set to true to turn on keep alives.
     */
    public synchronized void setKeepAliveEnabled(boolean enabled) {
        if (!enabled) {
            for (Iterator i = _sessions.iterator(); i.hasNext();) {
                HTTPClient session = (HTTPClient) i.next();
                i.remove();
                session.close();
            }
            _keepAlive = false;
        } else {
            _keepAlive = true;
        }
    }

    private synchronized void cleanSessionCache() {
        boolean ssl = TRANSPORT_TYPE_HTTPS == _transport;

        for (Iterator i = _sessions.iterator(); i.hasNext();) {
            HTTPClient session = (HTTPClient) i.next();
            if (!session.isCompatible(ssl, _server, getPort(), _authInfo, _useTLS) ||
                    session.isTimedOut()) {
                i.remove();
                session.close();
            }
        }

        /*
           *  When there are no sesions left in the cache, the thread
           *  clear the thread reference so that the watcher thread
           *  will end itself.
           */
        if (0 == _sessions.size()) {
            _watcherThread = null;
        }
    }

    /*
      *  Get a session, preferrably one from the cache, but if none are
      *  available, make a new one.
      */

    private synchronized HTTPClient getSession() throws IOException {
        boolean ssl = TRANSPORT_TYPE_HTTPS == _transport;
        boolean reserved = STYLE_HOSTSEQUIV == _style;
        /*
           * If server cannot be resolved throw an exception
           */
        if (_server == null) {
            try {
                _server = InetAddress.getByName(_serverName);
            } catch (UnknownHostException uhe) {
                throw new IOException("The host " +
                        _serverName + "could not be resolved");
            }
        }

        /*
           *  First look through the cache
           */
        while (_sessions.size() > 0) {
            HTTPClient session = (HTTPClient) _sessions.remove(0);
            if (!session.isTimedOut() &&
                    session.isCompatible(ssl, _server, getPort(), _authInfo, _useTLS)) {
                return session;
            } else {
                session.close();
            }
        }

        return new HTTPClient(ssl, _server, getPort(), reserved, _timeOut, _authInfo, _useTLS, _useSSL);
    }

    /*
      *  Place an open session in the cache for later use.
      */

    private synchronized void cacheSession(HTTPClient session) {
        _sessions.add(session);

        /*
           *  Create a thread that periodically cleans up old sessions
           *  which have timed out or become incompatible with the
           *  current settings.
           */
        if (null == _watcherThread) {
            _watcherThread = new Thread(new Runnable() {
                public void run() {
                    Thread me = Thread.currentThread();
                    while (_watcherThread == me) {
                        cleanSessionCache();
                        if (_watcherThread != me) {
                            break;
                        }
                        try {
                            Thread.sleep(5000);
                        } catch (InterruptedException e) {
                            ;
                        }
                    }
                }
            });
        }
    }

    /**
     * Set the TCP port used for API invocations on the server. <p>
     * Following are the default ports on the
     * storage system used for Data ONTAP API communication: <p>
     * <ul>
     *  <li>HTTP transport  - 80
     *  <li>HTTPS transport - 443
     * </ul>
     * <p>
     * Following are the default ports on the OnCommand Unified Manager 5.2 or earlier servers for
     * API communication: <p>
     * <ul>
     *  <li>HTTP transport  - 8088
     *  <li>HTTPS transport - 8488
     * </ul>
     * <p>
     * Following is the only supported port on the OnCommand Unified Manager 6.0 or later servers for
     * API communication: <p>
     * <ul>
     *  <li>HTTPS transport - 443
     * </ul>
     * <p>
     * Use the above port numbers as input to this API if the server has
     * default port settings.
     *
     * @param    tcpPort        The TCP server port to use.
     */
    public synchronized void setPort(int tcpPort) {
        _port = tcpPort;
    }

    /**
     * Get the TCP port used for API invocations on the server.
     */
    public int getPort() {
        if (0 != _port) {
            return _port;
        }
        switch (_type) {
            case SERVER_TYPE_NETCACHE:
                return TCP_PORT_NETCACHE;
            case SERVER_TYPE_AGENT:
                return TCP_PORT_AGENT;
            case SERVER_TYPE_DFM:
                return TCP_PORT_DFM;
            case SERVER_TYPE_OCUM:
                return TCP_PORT_OCUM;
            default:
            case SERVER_TYPE_FILER:
                return TCP_PORT_FILER;
        }
    }

    /**
     * Set the connection timeout value in seconds for the current
     * server context. <P>
     * Sets a connection timeout for the following actions: <P>
     * Establishing a connection with the server. <P>
     * Reading data from the server socket used for API communication. <P>
     * Note: Timeout value cannot be set for RPC authentication style.
     *
     * @param    timeout        The timeout value in seconds and should be
     * zero or a positive number. <P>
     * If it is non-zero, the server uses non-blocking
     * socket connection. If it is zero, the server uses
     * blocking connection, which is the default behavior.
     */
    public synchronized void setTimeout(int timeout) {
        if (timeout < 0) {
            throw new IllegalArgumentException(
                    "Invalid timeout value: " + timeout);
        }
        _timeOut = timeout;
    }

    /**
     * Get the connection timeout value in seconds for the current
     * server context. <P>
     */
    public int getTimeout() {
        return _timeOut;
    }

    /**
     * Get the server type, which is one of the SERVER_TYPE_ values.
     */
    public int getServerType() {
        return _type;
    }

    /*
      *  Get the appropriate URI suffix for the current sever type
      */

    private String getURISuffix() {
        switch (_type) {
            case SERVER_TYPE_NETCACHE:
                return NETCACHE_REQUEST_URI;
            case SERVER_TYPE_AGENT:
                return AGENT_REQUEST_URI;
            case SERVER_TYPE_DFM:
                return DFM_REQUEST_URI;
            case SERVER_TYPE_OCUM:
                return DFM_REQUEST_URI;
            default:
            case SERVER_TYPE_FILER:
                return FILER_REQUEST_URI;
        }
    }

    /**
     * Set the URI and TCP port appropriate for the given server type.
     *
     * @param    serverType        The type of API server.
     * <p>
     * Valid server types are:
     * <ul>
     *  <li>NaServer.SERVER_TYPE_FILER - For connecting to Data ONTAP storage systems.
     *  <li>NaServer.SERVER_TYPE_DFM   - For connecting to OnCommand Unified Manager server 5.2 or earlier.
     *  <li>NaServer.SERVER_TYPE_OCUM  - For connecting to OnCommand Unified Manager server 6.0 or later.
     * </ul>
     * <p>
     * The default server type on NaServer construction is SERVER_TYPE_FILER.
     * @exception IllegalArgumentException    if server type is not valid.
     */
    public synchronized void setServerType(int serverType) {
        /*
           * Switch on the known server types.  Unknown types will cause
           * an exception.
           */
        switch (serverType) {
            case SERVER_TYPE_FILER:
            case SERVER_TYPE_NETCACHE:
            case SERVER_TYPE_AGENT:
            case SERVER_TYPE_DFM:
                _type = serverType;
                return;
            case SERVER_TYPE_OCUM:
                _type = serverType;
                _transport = TRANSPORT_TYPE_HTTPS;
                _port = TCP_PORT_OCUM;
                return;
            default:
                throw new IllegalArgumentException(
                        "Unknown API server type: " + serverType);
        }
    }

    /**
     * Get the transport type, which is one of the TRANSPORT_TYPE_ values.
     */
    public int getTransportType() {
        return _transport;
    }

    /**
     * Set the transport type. <P>
     * After setting the transport type, call the setPort()
     * to set the port number used for that transport type.
     *
     * @param    transportType    The type of transport.
     * <p>
     * Valid transport types are: 
     * <ul>
     *  <li>NaServer.TRANSPORT_TYPE_HTTP
     *  <li>NaServer.TRANSPORT_TYPE_HTTPS
     * </ul>
     * <p>
     * TRANSPORT_TYPE_HTTP is the default transport type for SERVER_TYPE_FILER and SERVER_TYPE_DFM.
     * For SERVER_TYPE_OCUM, TRANSPORT_TYPE_HTTPS is the only supported transport type and is set by default.
     * @exception IllegalArgumentException    if transport is not valid.
     */
    public synchronized void setTransportType(int transportType) {
        /*
           * Switch on the known transport types.  Unknown types will cause
           * an exception.
           */
        switch (transportType) {
            case TRANSPORT_TYPE_HTTP:
		if(_type == SERVER_TYPE_OCUM) {
                    throw new IllegalArgumentException(
                        "'SERVER_TYPE_OCUM' does not support 'TRANSPORT_TYPE_HTTP'");
		} else {
                    _transport = transportType;
		}
                break;
            case TRANSPORT_TYPE_HTTPS:
                _transport = transportType;
                break;
            default:
                throw new IllegalArgumentException(
                        "Unknown API transport type: " + transportType);
        }
        if (_transport != TRANSPORT_TYPE_HTTPS) {
            _authInfo.disableCBA();
        }
        // Try to check whether TLS protocol can be enabled for SSL
        // communication from the lookup table that is set by a previous
        // instance of NaServer that has connected to the same server IP.
        // If the value is not set (we are the first NaServer instance
        // to connect), we try to enable TLS and update the lookup table.
        // If the connection fails, then we try to disable TLS and update the
        // lookup table.
        if (_transport == TRANSPORT_TYPE_HTTPS) {
            Boolean tlsInUse = (Boolean) _tlsInUse.get(_serverName);
            if (tlsInUse == null) {
                tlsInUse = Boolean.TRUE;
                _tlsInUse.put(_serverName, tlsInUse);
            }
            _useTLS = tlsInUse.booleanValue();
        }
    }

    /**
     * Set the login and password used for authenticating when
     * an API is invoked.
     *
     * @param    login        The login name (often "root").
     * @param    password    The password.
     */
    public void setAdminUser(String login, String password) {
        _login = login;
        _password = password;
    }

    /**
     * Set the authentication style for subsequent API authentications.
     *
     * @param    style        The style, must be one of:
     * NaServer.STYLE_LOGIN_PASSWORD
     * NaServer.STYLE_HOSTSEQUIV
     * NaServer.STYLE_CERTIFICATE
     * NaServer.STYLE_RPC <P>
     * NaServer.STYLE_RPC and NaServer.STYLE_HOSTSEQUIV styles 
     * are applicable for Data ONTAP APIs only. <P>
     * NaServer.STYLE_CERTIFICATE is currently applicable for 
     * DataFabric Manager server API communication only. <P>
     *
     * If style = NaServer.STYLE_CERTIFICATE, you can use certificates to 
     * authenticate clients who attempt to connect to a server without the need 
     * of username and password. This style will internally set the transport type
     * to HTTPS (TRANSPORT_TYPE_HTTPS). <P>
     *
     * Verification of the server's certificate is required in order to 
     * properly authenticate the identity of the server.
     * Server certificate (with hostname) verification will be enabled by default using this style. <P>
     * You can disable server certificate (with hostname) verification using disableServerCertVerification() and 
     * and you can disable only hostname verification using disableHostnameVerification(). <P>
     *
     * You can use setKeyStore() to specify the keystore containing 
     * client certificates and  setTrustStore() to specify the 
     * truststore containing the certificates of the trusted 
     * Certificate Authorities required for authenticating the server.
     *
     * @exception    IllegalArgumentException    if style is not valid.
     */
    public void setStyle(int style)
    {
        if (STYLE_LOGIN_PASSWORD != style && STYLE_HOSTSEQUIV != style &&
                STYLE_RPC != style && STYLE_CERTIFICATE != style) {
            throw new IllegalArgumentException("Invalid style: "+style);
        }
        if (STYLE_CERTIFICATE == style) {
            setTransportType(TRANSPORT_TYPE_HTTPS);
            _authInfo.enableCBA();
            enableServerCertVerification();
        }
        else {
            _authInfo.disableCBA();
        }
        _style = style;
    }

    /**
     * Set the snoop (debugging) level.
     *
     * @param    level        0 (the default) means no debugging.
     * -1 means to print the raw XML input and output.
     */
    public void setSnoop(int level) {
        _snoop = level;
        HTTPClient.snoop = level;
    }

    /**
     * Get the current authentication style.
     *
     * @return Integer value (possible values: NaServer.STYLE_LOGIN_PASSWORD,
     * NaServer.STYLE_HOSTSEQUIV, NaServer.STYLE_RPC)
     */
    public int getStyle() {
        return _style;
    }


    /**
     * Invoke a single API request.
     *
     * @param    xi        The API element.
     * @return Output results of the API.
     * @exception IOException    if there is a communication problem
     * @exception NaAuthenticationException    if the login was not accepted
     * @exception NaAPIFailedException    if the server returned a failure
     * result.
     * @exception NaProtocolException    if there is a protocol problem
     */
    public NaElement invokeElem(NaElement xi) throws IOException,
            NaAuthenticationException, NaAPIFailedException, NaProtocolException {
        String xmlStr = invokeElemForXML(xi);

        NaElement xo;

        try {
            xo = NaParser.parseTree(new StringReader(xmlStr));
        } catch (IOException e) {
            throw new NaError("I/O error during XML parsing", e);
        } catch (SAXException e) {
            throw new NaProtocolException("Malformed XML", e);
        }


        /*
           * We expect a <netapp> element
           */
        if (null == xo || !xo.getName().equals("netapp")) {
            throw new NaProtocolException("Output from server doesn't "
                    + "contain <netapp> element.");
        }

        if (xo.getChildren().size() > 1) {
            if (!(xo.getChildren().size() == 2 
	          && xo.getChildByName("trace-output") != null)) {
	     return xo;
	    }
	}

        NaElement results = xo.getChildByName("results");

        /*
           *  If the API succeeded, return the results now
           */
        if ("passed".equals(results.getAttr("status"))) {
            return results;
        }

        /*
           *  Otherwise, something went wrong, and we need to build and
           *  throw an NaAPIFailedException.
           */
        int errno;
        String s = results.getAttr("errno");
        if (null == s) {
            /*
                *  This is to handle the rare case when the errno is missing,
                *  which only occurs in Ontap 6.3, as far as I know.
                */
            errno = NaErrno.EINTERNALERROR;
        }
        try {
            errno = Integer.parseInt(s);
        } catch (NumberFormatException e) {
            throw new NaProtocolException("Malformed errno code: " + s, e);
        }
        throw new NaAPIFailedException(errno, results.getAttr("reason"));
    }

    /**
     * Invoke a single ONTAPI API.
     *
     * @throws IOException
     * @throws NaAuthenticationException
     * @throws NaAPIFailedException
     * @throws NaProtocolException
     * @param    xi        The API element.
     * @return Output results of the API in XML format.
     */
    public String invokeElemForXML(NaElement xi) throws IOException,
            NaAuthenticationException, NaAPIFailedException, NaProtocolException {

        if(xi.getName().equals("netapp")) {
            multipleApis = true;
            if(xi.getAttr("version") != null && xi.getAttr("xmlns") != null) {
                apiAttributeSet = true;
            }
        }

        String xml = xi.toString();
        String body;

        if (STYLE_RPC == _style) {
            body = invokeRPC(xml);
        } else {
            body = invokeHTTP(xml);
        }

        /* Show the output in raw <XML> form */
        if (_snoop == -1) {
            setRawXMLOutput(body);
            System.out.println("OUTPUT:\n" + body);
        }
        if (_snoop > 1) {
            System.err.println("BODY =====\n" + body +
                    "\n===== END BODY");
        }
        if (0 == body.length()) {
            throw new NaProtocolException("body of server response empty");
        }

        /*
        * The parser doesn't like the <!DOCTYPE> entry which
        * contains a relative URI.  I wasn't able to figure out how
        * to tell the parser what our parent URL/hostname/whatever was,
        * so that it could handle the relatvie URI,
        * so to finesse this, we just remove the <!DOCTYPE> tag entirely.
        *
        * It's also possible that the <!DOCTYPE> doesn't appear at all,
        * in which case we've got a fairly degenerate situation (like
        * a 404 Not Found response)
        */
        int pos1 = body.indexOf("<!DOCTYPE");
        int pos2 = body.indexOf("<netapp");

        /*
        * Maybe we failed to find a <!DOCTYPE>
        *
        * This could be an authentication error.  It could also be that
        * the server doesn't have a clue (404 Not Found), which might be
        * the case if the server is listening but is not a ZAPI server.
        *
        * We err on the side of hinting to the client to try again.
        */
        if (pos1 < 0) {
            throw new NaProtocolException("server response did not " +
                    "contain <!DOCTYPE>");
        }

        /*
        * Maybe we just failed to find <netapp>
        */
        if (pos2 < 0) {
            /*
            * We found a <!DOCTYPE>, but there's no netapp directive.
            * We'll try to return whatever follows the <!DOCTYPE>
            */
            pos2 = body.indexOf("<", pos1);

            if (pos2 < 0) {
                throw new NaProtocolException("Output from server doesn't " +
                        "contain <netapp> element.");
            }

            throw new NaProtocolException("Unexpected tag found: " +
                    body.substring(0, pos1) + body.substring(pos2));
        }

        /*
        * strip <!DOCTYPE> from body
        */
        StringBuffer auxStringBuffer = new StringBuffer();
        auxStringBuffer.append(body.substring(0, pos1)).append(body.substring(pos2));

        if (_snoop > 1) {
            System.err.println("MODIFIED BODY =====\n" + auxStringBuffer.toString() +
                    "\n===== END MODIFIED BODY");
        }
        return auxStringBuffer.toString();
    }

    /**
     * Invoke the API using HTTP or HTTPS
     */
    private String invokeHTTP(String xml) throws IOException,
            NaAuthenticationException, NaProtocolException {
        String XML_REQUEST_BEGIN = null;
        /*
           *  Build the request object that represent the API invocation
           */
        HTTPRequest request = new HTTPRequest(HTTPRequest.METHOD_POST,
                getURISuffix(), HTTPRequest.VERSION_1_0);
        byte[] content;

        //Added for vfiler-tunneling
        String vfiler_tunnel = "";
        if (_vfiler != null) {
            vfiler_tunnel = " vfiler=" + "'" + _vfiler + "'";

        }
        String nmsdkAppReq = "";
        if(nmsdkAppName != null) {
            nmsdkAppReq = " nmsdk_app='" + nmsdkAppName + "'";
        }

        try {
            switch (_type) {
                case SERVER_TYPE_DFM:
                    XML_REQUEST_BEGIN = DFM_XML_REQUEST;
                    break;
                case SERVER_TYPE_OCUM:
                    XML_REQUEST_BEGIN = DFM_XML_REQUEST;
                    break;
                case SERVER_TYPE_AGENT:
                    XML_REQUEST_BEGIN = AGENT_XML_REQUEST;
                    break;
                case SERVER_TYPE_FILER:
                    XML_REQUEST_BEGIN = FILER_XML_REQUEST;
                    break;
            }

            if(apiAttributeSet) {
                content = (XML_REQUEST_BEGIN + xml + "\r\n").getBytes("UTF-8");
            } else {
                if(multipleApis) {
                   xml = xml.replaceFirst("<netapp>", "");
                   xml = xml.replaceAll("</netapp>", "");
                }
                content = (XML_REQUEST_BEGIN + "<netapp version='" +
                    _major + "." + _minor + "' xmlns='" + FILER_NAMESPACE + "'"
                    + vfiler_tunnel
                    + " nmsdk_version='" + NMSDK_VERSION + "'"
                    + " nmsdk_platform='" + NMSDK_PLATFORM + "'"
                    + " nmsdk_language='" + NMSDK_LANGUAGE + "'"
                    + nmsdkAppReq
                    + ">"
                    + xml + "</netapp>\r\n").getBytes("UTF-8");
            }

        } catch (UnsupportedEncodingException e) {
            throw new NaError("Failed UTF-8 encoding", e);
        }

        /* Show the INPUT in raw <XML> form */
        if (_snoop == -1) {
            System.out.println("INPUT:\n" + XML_REQUEST_BEGIN + "<netapp" + vfiler_tunnel + " xmlns='" +
                    FILER_NAMESPACE + " ' " + " version='" + _major +
                    "." + _minor + "'"
                    + " nmsdk_version='" + NMSDK_VERSION + "'"
                    + " nmsdk_platform='" + NMSDK_PLATFORM + "'"
                    + " nmsdk_language='" + NMSDK_LANGUAGE + "'"
                    + nmsdkAppReq
                    + ">" + xml + "</netapp>\r\n");
        }

        request.setContents(content);

        //Add a cookie header, if set
        if (cookieValue != null) {
	        request.setHeader("Cookie", cookieValue);
        }

        request.setHeader("Content-Length", "" + content.length);
        if (_style != STYLE_HOSTSEQUIV)
            request.setBasicAuthorization(_login, _password);
        if (_keepAlive) {
            request.setHeader("Connection", "Keep-Alive");
        }

        HTTPResponse response;

        while (true) {
            /*
                *  Get a connection to the server
                */
            HTTPClient session = getSession();
            boolean fresh = (0 == session.getUseCount());

            if (_snoop > 0) {
                System.err.println((fresh) ? "*** Using new connection" :
                        "*** Reusing existing connection");
            }

            try {
                /*
                     *  Do the actual HTTP transaction
                     */
                response = session.doRequest(request);
            } catch (IOException e) {
                session.close();
                // If the connection over TLS fails, then update the lookup table
                // and reconnect using SSLv3.
                if (_useTLS && _useSSL && e.getCause() instanceof javax.net.ssl.SSLException) {
                    String msg = e.getMessage();
                    if ((msg != null) && (msg.indexOf("bad_record_mac") != -1)) {
                        _tlsInUse.put(_serverName, Boolean.FALSE);
                        _useTLS = false;
                        if (_snoop > 0) {
                            System.err.println("*** Failed TLS attempt for server " + _serverName + ", trying SSLv3 ..");
                        }
                        continue;
                    }
                }
                if (fresh) {
                    throw e;
                } else {
                    /*
                          *  The connection was reused, so try
                          *  again.
                          */
                    if (_snoop > 0) {
                        System.err.println("*** Failed attempt (" +
                                e + ") trying again");
                    }
                    continue;
                }
            } catch (HTTPException e) {
                session.close();
                throw new NaProtocolException("HTTP protocol error", e);
            }

            /*
                *  If this session is reusable, save it for later
                */
            if (_keepAlive && response.canKeepAlive()) {
                if (_snoop > 0) {
                    System.err.println("*** Saving connection for future use");
                }
                cacheSession(session);
            } else {
                if (_snoop > 0) {
                    System.err.println("*** Closing non-reusable connection");
                }
                session.close();
            }
            break;
        }

        /*
           *  Check the HTTP status
           */
        if (401 == response.getStatusCode()) {
            throw new NaAuthenticationException("Authorization failed");
        } else if (200 != response.getStatusCode()) {
            throw new NaProtocolException("Unexpected HTTP response: " +
                    response.getStatusCode() + " " + response.getReasonPhrase());
        }


        /*
           *  Convert the message body to a giant string
           */
        try {
            return new String(response.getContents(), "UTF-8");
        } catch (UnsupportedEncodingException e) {
            throw new NaError("Failed UTF-8 decoding", e);
        }
    }

    /**
     * Parses the given API request in the form of XML and builds an NaElement of it.
     *
     * @param args XML input string
     * @return NaElement of the API request.
     */
    public NaElement getXMLParseInput(String args) throws IOException, SAXException, NaProtocolException {
        NaElement xo;
        int argcnt = 0;
        try {
            xo = NaParser.parseTree(new StringReader(args));
        } catch (IOException e) {
            throw new NaError("I/O error during XML parsing", e);
        } catch (SAXException e) {
            throw new NaProtocolException("Malformed XML", e);
        }
        return xo;
    }

    /*
      * setRawXMLOutput
      * Save the RawXML content.
      *
      */

    private void setRawXMLOutput(String XML) {
        _rawXMLOutput = XML;
    }


    /**
     * Invoke the API using RPC
     */
    private String invokeRPC(String xml) throws IOException,
            NaAuthenticationException, NaProtocolException {
        NaInvokable invokable;

        try {
            Class invokeClass = Class.forName("netapp.manage.NaRPCTarget");
            invokable = (NaInvokable) invokeClass.newInstance();
        } catch (ClassNotFoundException e) {
            throw new NaError("Failed to load RPC library", e);
        } catch (InstantiationException e) {
            throw new NaError("Failed to instantiate RPC library", e);
        } catch (IllegalAccessException e) {
            throw new NaError("Illegal access loading RPC library", e);
        }

        invokable.setHostname(_serverName);

        xml = "<netapp version='" + _major + "." + _minor + "' xmlns='" +
                FILER_NAMESPACE + "'>" + xml + "</netapp>\r\n";

        return invokable.invoke(xml);
    }

    /**
     * get a string representation of the transport type
     */
    private String getTransportString() {
        switch (_transport) {
            case TRANSPORT_TYPE_HTTP:
                return "HTTP";
            case TRANSPORT_TYPE_HTTPS:
                return "HTTPS";
            default:
                return Integer.toString(_transport);
        }
    }

    /**
     * get a string representation of the authentication style
     */
    private String getAuthStyleString() {
        switch (_style) {
            case STYLE_LOGIN_PASSWORD:
                return "password";
            case STYLE_HOSTSEQUIV:
                return "host-equiv";
            case STYLE_RPC:
                return "RPC";
            case STYLE_CERTIFICATE:
                return "CERTIFICATE";
            default:
                return Integer.toString(_style);
        }
    }

    /*
     *
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();

        sb.append("[ ");
        sb.append("server=" + _server.getHostName());
        sb.append(",");
        sb.append("uri=" + getURISuffix());
        sb.append(",");
        sb.append("login=" + _login);
        sb.append(",");
        sb.append("password=(hidden)");
        sb.append(",");
        sb.append("auth-style="+getAuthStyleString());
		sb.append(",");
		sb.append("protocol="+getTransportString());
		sb.append(",");
		sb.append("port="+getPort());
		sb.append(",");
		sb.append("version="+_major+"."+_minor);
		sb.append(" ]");

		return sb.toString();
	}

	/**
     * Set the name of the vfiler on which the Data ONTAP API commands need
     * to be invoked.
     */
	public synchronized  void setVfilerTunneling(String vfiler)
	{
		if (_major >= 1) {
			if (_minor >= 7) {
				_vfiler = vfiler;
				return;
			}
		}
		throw new IllegalArgumentException(
		 "ONTAPI version must be at least 1.7 to send API to a vfiler");
	}
	
    /**
     * Adds a cookie header to the outgoing HTTP request.
     * This menthod is used by NetApp's internal products.
     *
     * @param cookieValue Value of the cookie header
     */
    public void setCookie(String cookieValue)
    {
		this.cookieValue = cookieValue;
    }

    /**
     * Gets the cookie header that will be added to the outgoing HTTP request.
     * This menthod is used by NetApp's internal products.
     *
     * @return Value of the cookie header
     */
    public String getCookie()
    {
        return this.cookieValue;
    }

	/**
	* Sets the name of the vserver to which a Cluster API need to be tunneled
	* from a Cluster Management Interface.
	* @param   vserver         The name of the vserver.
	*/
	public synchronized  void setVserver(String vserver)
	{
		if(vserver == null) {
			throw new NullPointerException("vserver name must not be null");
		}

		if (_major >= 1 && _minor >= 15) {
			_vfiler = vserver;
			return;
		}

		throw new IllegalArgumentException(
		"ONTAPI version must be at least 1.15 to send API to a vserver");	
	}

	/**
	* Gets the name of the vserver to which a Cluster API need to be tunneled
	* from a Cluster Management Interface.
	* @return	name of the vserver	
	*/
	public synchronized String getVserver()
	{
		return _vfiler;
	}

	/**
	* Sets the location of the keystore file where the client certificate 
	* and the key reside.
	* @param   keyStore         The location of the keystore file.
	* @param   keyStorePasswd   The password to access the keystore file.
	*
	* @exception	IllegalArgumentException	if keyStore is null.
	*/
	public void setKeyStore(String keyStore, String keyStorePasswd)
	{
		setKeyStore(keyStore, keyStorePasswd, null);
	}

	/**
	* Sets the location of the keystore file where the client certificate 
	* and the key reside.
	* @param   keyStore         The location of the keystore file.
	* @param   keyStorePasswd   The password to access the keystore file.  
	* @param   keyPasswd        The password to access the individual key
	*                           in the keystore file. This is used if the key 
	*                           in the keystore file have a different password 
	*                           than the keystore file.
	*
	* @exception	IllegalArgumentException	if keyStore is null.
	*/
	public void setKeyStore(String keyStore, String keyStorePasswd,
			String keyPasswd)
	{
		if (keyStore == null) {
			throw new IllegalArgumentException("KeyStore file cannot be null");
		}
		_authInfo.setKeyStore(keyStore, keyStorePasswd, keyPasswd);
	}

	/**
	* Sets the keystore type for client certificates. 
	* @param    keyStoreType    The type of the keystore file.
	*           Should be one of the NaServer.KEYSTORE_TYPE_* values. <br>
	*           The default type is KEYSTORE_TYPE_JKS (Java Key Store).
	*
	* @exception	IllegalArgumentException	if keyStoreType is invalid.
	*/
	public void setKeyStoreType(int keyStoreType)
	{
		switch (keyStoreType) {
			case KEYSTORE_TYPE_JKS: 
				_authInfo.setKeyStoreType("JKS"); 
				break;
			case KEYSTORE_TYPE_PKCS12:
				_authInfo.setKeyStoreType("PKCS12");
				break;
			default:
				throw new IllegalArgumentException(
				"Invalid keystore type: "+ keyStoreType);
		}
	}

	/**
	* Enables the server certificate verification by the client. 
	* This method will internally enable hostname verification.
	* The server certificate verification is enabled by default when 
	* the authentication style is set to STYLE_CERTIFICATE. <P>
	* Use setTrustStore() method to specify the truststore containing the 
	* certificates of the trusted Certificate Authorities required for 
	* authenticating the server. 
	* You can disable hostname verification using disableHostnameVerification(). <P>
	* You need to set the transport type to HTTPS (TRANSPORT_TYPE_HTTPS) before 
	* calling this method.
	* 
	* @exception	IllegalStateException	if transport type is not set to HTTPS.
	*/
	public void enableServerCertVerification()
	{
		if (_transport != TRANSPORT_TYPE_HTTPS) {
			throw new IllegalStateException(
				"Server certificate verification can only be enabled for HTTPS transport");
		}
		_authInfo.enableServerCertVerification();
	}

	/**
	* Disables the server certificate verification by the client.
	* This method will internally disable hostname verification. <P>
	* You need to set the transport type to HTTPS (TRANSPORT_TYPE_HTTPS) before 
	* calling this method.
	* 
	* @exception	IllegalStateException	if transport type is not set to HTTPS.
	*/
	public void disableServerCertVerification()
	{
		if (_transport != TRANSPORT_TYPE_HTTPS) {
			throw new IllegalStateException(
				"Server certificate verification can only be disabled for HTTPS transport");
		}
		_authInfo.disableServerCertVerification();
	}

	/**
	* Determines whether Server certificate verification is enabled by the client.
	*/
	public boolean isServerCertVerificationEnabled()
	{
		return _authInfo.isServerCertVerificationEnabled();
	}

	/**
	 * Enables Hostname verification by the client during server 
	 * certificate verification.
	 * Hostname verification is enabled by default when the authentication 
	 * style is set to STYLE_CERTIFICATE.<p>
	 * Hostname verification ensures that the hostname to which the client 
	 * connects matches the hostname (CN name) in the certificate that the 
	 * server sends back as part of the SSL connection. <p>
	 * For hostname verification, the application must specify the hostname (CN name) 
	 * instead of the IP address it is connecting to the server in NaServer constructor. <p>
	 * Server certificate verification must be enabled before using this method.
	 *
	 * @exception	IllegalStateException	if server certificate verification is not enabled.
	 */
	public void enableHostnameVerification()
	{
		if (!_authInfo.isServerCertVerificationEnabled()) {
			throw new IllegalStateException(
				"Server certificate verification is not enabled");
		}
		_authInfo.enableHostnameVerification();
	}

	/**
	 * Disables Hostname verification by the client during server 
	 * certificate verification.
	 *
	 * @exception	IllegalStateException	if server certificate verification is not enabled.
	 */
	public void disableHostnameVerification()
	{
		if (!_authInfo.isServerCertVerificationEnabled()) {
			throw new IllegalStateException(
				"Server certificate verification is not enabled");
		}
		_authInfo.disableHostnameVerification();
	}

	/**
	 * Determines whether Hostname verification is enabled by the client during 
	 * server certificate verification.
	 */
	public boolean isHostnameVerificationEnabled()
	{
		return _authInfo.isHostnameVerificationEnabled();
	}
	
	/**
	* Sets the default location of the truststore containing the certificates of the 
	* trusted Certificate Authorities required for authenticating the server. <P>
	* @param   trustStore       The location of the truststore file.
	*
	* @exception	IllegalArgumentException	if trustStore is null.
	*/  
	public void setTrustStore(String trustStore)
	{
		if (trustStore == null) {
			throw new IllegalArgumentException("TrustStore file cannot be null");
		}
		_authInfo.setTrustStore(trustStore);
	}

	/**
	 *  Enables TLS protocol for use on connection over HTTPS Transport.
	 *  By default, TLS protocol is enabled.
	 */
	public void enableTLS()
	{
		if (_transport != TRANSPORT_TYPE_HTTPS) {
			throw new IllegalStateException(
				"TLS protocol can only be enabled for HTTPS transport");
		}
		_useTLS = true;
	}

	/**
	 *  Disables TLS protocol for use over HTTPS Transport.
	 *  By default, TLS protocol is enabled.
	 */
	public void disableTLS()
	{
		if (_transport != TRANSPORT_TYPE_HTTPS) {
			throw new IllegalStateException(
				"TLS protocol can only be disabled for HTTPS transport");
		}
		_useTLS = false;
	}

	/**
	 *  Determines whether TLS protocol is enabled for use over HTTPS Transport.
	 */
	public boolean isTLSEnabled()
	{
		return _useTLS;
	}

	/**
	 *  Enables SSLv3 protocol for use over HTTPS Transport.
	 *  By default, SSLv3 protocol is disabled.
	 */
	public void enableSSLv3()
	{
		if (_transport != TRANSPORT_TYPE_HTTPS) {
			throw new IllegalStateException(
				"SSL protocol can only be disabled for HTTPS transport");
		}
		_useSSL = true;
	}
	
	/**
	 *  Disables SSLv3 protocol for use over HTTPS Transport.
	 *  By default, SSLv3 protocol is disabled.
	 */
	public void disableSSLv3()
	{
		if (_transport != TRANSPORT_TYPE_HTTPS) {
			throw new IllegalStateException(
				"SSL protocol can only be disabled for HTTPS transport");
		}
		_useSSL = false;
	}
	
	
	/**
	 *  Determines whether SSLv3 protocol is enabled for use over HTTPS Transport.
	 */
	public boolean isSSLv3Enabled()
	{
		return _useSSL;
	}

	/**
	 * Sets the name of the client application.
	 *
	 * @param appName Client application name
	 */
	public static void setApplicationName(String appName)
	{
		nmsdkAppName = appName;
	}

	/**
	 * Returns the name of the client application.
	 *
	 * @return Name of the client application
	 */
	public static String getApplicationName()
	{
		return nmsdkAppName;
	}


	/**
	 * Returns the platform information. This method is used internally.
	 *
	 * @return platform information
	 */
	private static String getPlatformInfo()
	{
		String osType = System.getProperty("os.name");
		String osInfo = "";

		if (osType.startsWith("Windows")) {
			return getWindowsPlatformInfo();
		} else {
			try {
				Runtime runtime = Runtime.getRuntime();
				Process process;
				BufferedReader bufferedReader;
				String[] cmd;
				String osName = "";
				String processor = "";

				if (osType.equals("Linux")) {
					boolean foundOsName = true;
					osName = "Linux";
					File file = new File("/etc/SuSE-release");
					if (file.exists()) {
						cmd = new String[] {"sh", "-c", "head -n 1 /etc/SuSE-release && uname -p"};
					} else {
						file = new File("/etc/redhat-release");
						if (file.exists()) {
							cmd = new String[] {"sh", "-c", "head -n 1 /etc/redhat-release && uname -p"};
						} else {
							cmd = new String[] {"sh", "-c", "uname -p"};
							foundOsName = false;
						}
					}
					process = runtime.exec(cmd);
					process.waitFor();
					if (process.exitValue() != 0) {
						return osType;
					}
					bufferedReader = new BufferedReader(new InputStreamReader(process.getInputStream()));
					if (foundOsName) {
						osName = bufferedReader.readLine();
					}
					processor = bufferedReader.readLine();
					bufferedReader.close();
					String regEx = "(.*?) \\(.*?\\)";
					Pattern pattern = Pattern.compile(regEx);
					Matcher matcher = pattern.matcher(osName);
					if (matcher.find()) {
						osName = matcher.group(1);
					}
					osInfo = osName + " " + processor;
				} else if (osType.equals("SunOS")) {
					cmd = new String [] {"sh", "-c", "uname -srp && isainfo -b"};
					process = runtime.exec(cmd);
					process.waitFor();
					if (process.exitValue() != 0) {
						return osType;
					}
					bufferedReader = new BufferedReader(new InputStreamReader(process.getInputStream()));
					String unameInfo = bufferedReader.readLine();
					String bitInfo = bufferedReader.readLine() + "-bit";
					osInfo = unameInfo + " " + bitInfo;
					bufferedReader.close();
				} else if (osType.equals("HP-UX")) {
					cmd = new String [] {"sh", "-c", "uname -srm"};
					process = runtime.exec(cmd);
					process.waitFor();
					if (process.exitValue() != 0) {
						return osType;
					}
					bufferedReader = new BufferedReader(new InputStreamReader(process.getInputStream()));
					osInfo = bufferedReader.readLine();
					bufferedReader.close();
				} else if (osType.equals("AIX")) {
					cmd = new String [] {"sh", "-c", "uname -sp && oslevel && getconf KERNEL_BITMODE"};
					process = runtime.exec(cmd);
					process.waitFor();
					if (process.exitValue() != 0) {
						return osType;
					}
					bufferedReader = new BufferedReader(new InputStreamReader(process.getInputStream()));
					String unameInfo = bufferedReader.readLine();
					String osLevelInfo = bufferedReader.readLine();
					String bitInfo = bufferedReader.readLine();
					bufferedReader.close();
					String[] split = unameInfo.split("\\s+");
					osName = split[0];
					processor = split[1];
					osInfo = osName + " " + osLevelInfo + " " + processor + " " + bitInfo + "-bit";
				} else {
					osInfo = osType;
				}
			} catch (Exception e) {
				osInfo = osType;
			}
		}
		return osInfo;
	}

	/**
	 * Returns the platform information on windows. This method is used internally.
	 *
	 * @return platform information
	 */
	private static String getWindowsPlatformInfo() {
		final String osKeyName = "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion";
		final String osValueName = "ProductName";
		final String processorKeyName = "SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment";
		final String processorValueName = "PROCESSOR_ARCHITECTURE";
		String osName = getWindowsRegistryValue(osKeyName, osValueName);
		String processor;
		String osInfo;

		if (osName != null) {
			processor = getWindowsRegistryValue(processorKeyName, processorValueName);
				if (processor != null) {
					osInfo = osName + " " + processor;
				} else {
					osInfo = osName;
				}
		} else {
			osInfo =  System.getProperty("os.name");
		}
		return osInfo;
	}

	/**
	 * Returns the registry value data for the specied key and value names. This method is used internally.
	 *
	 * @return registry value data
	 */
	//@SuppressWarnings({ "unchecked", "rawtypes" })
	private static String getWindowsRegistryValue(String keyName, String valueName) {
		final int ERROR_SUCCESS = 0;
		final int ERROR_CODE = 1;
		Integer hKey;
		byte[] value = null;
		final Integer HKEY_LOCAL_MACHINE = new Integer(0x80000002);
		final Integer KEY_READ = new Integer(0x20019);

		try {
			final Class userClass = Preferences.userRoot().getClass();
			Class openKeyParams[] = {int.class, byte[].class, int.class};
			Class closeKeyParams[] = {int.class};
			Class queryValueParams[] = {int.class, byte[].class};
			final Method openKeyMethod = userClass.getDeclaredMethod("WindowsRegOpenKey", openKeyParams);
			final Method closeKeyMethod = userClass.getDeclaredMethod("WindowsRegCloseKey", closeKeyParams);
			final Method queryValueMethod = userClass.getDeclaredMethod("WindowsRegQueryValueEx", queryValueParams);
			openKeyMethod.setAccessible(true);
			queryValueMethod.setAccessible(true);
			closeKeyMethod.setAccessible(true);

			Object openKeyParamValues[] = {HKEY_LOCAL_MACHINE, stringToByteArray(keyName), KEY_READ};
			int[] result = (int[]) openKeyMethod.invoke(null, openKeyParamValues);
			if (result[ERROR_CODE] != ERROR_SUCCESS) {
				return null;
			}
			hKey = new Integer(result[0]);
			Object queryValueParamValues[] = {hKey, stringToByteArray(valueName)};
			value = (byte[]) queryValueMethod.invoke(null, queryValueParamValues);
			Object closeKeyParamValues[] = {hKey};
			closeKeyMethod.invoke(null, closeKeyParamValues);
			if (value == null) {
				return null;
			}
		} catch (Exception e) {
			return null;
		}
		return new String(value).trim();
	}

	private static byte[] stringToByteArray(String str) {
		byte[] result = new byte[str.length() + 1];
		for (int i = 0; i < str.length(); i++) {
			result[i] = (byte) str.charAt(i);
		}
		result[str.length()] = 0; 
		return result;
	}
}
