/*
 * $Id: //depot/prod/zephyr/lufthansaair/src/java/netapp/manage/http/AuthInfo.java#1 $
 *
 * Copyright (c) 2010 NetApp, Inc.
 * All rights reserved.
 *
 * Certificate Based Authentication stuff.
 * This class is unsupported. It is for internal use only.
 */

package netapp.manage.http;

/**
 *  Class that contains information about keystore and tuststore details that are 
 *  required for certificate based authentication.
 */
public class AuthInfo
{
private String _keyStoreFile;
private String _keyStorePasswd;
private String _keyPasswd;
private String _keyStoreType;
private String _trustStoreFile;
private String _trustStoreType;
private boolean _isCBAEnabled;
private String _hostName;
private boolean _isServerCertVerificationEnabled;
private boolean _isHostnameVerificationEnabled;

    /**
     *  Constructor which will initialize keystore and truststore values.
    */
    public AuthInfo(String hostName)
    {
        _keyStoreFile = null;
        _keyStorePasswd = null;
        _keyPasswd = null;
        _keyStoreType = "JKS";
        _trustStoreFile = null;
        _trustStoreType = "JKS";
        _isCBAEnabled = false;
        _hostName = hostName;
        _isServerCertVerificationEnabled = false;
        _isHostnameVerificationEnabled = false;
    }

    /**
     *  Set the keystore details where the client certificate and the key resides.
     *
     *  @param    keyStoreFile    Location of the keystore file.
     *  @param    keyStorePasswd  Password to access the keystore file.
     *  @param    keyPasswd       Password to access the private key.
     */
    public void setKeyStore(String keyStoreFile, String keyStorePasswd, 
            String keyPasswd)
    {
        _keyStoreFile = keyStoreFile;
        _keyStorePasswd = keyStorePasswd;
        _keyPasswd = keyPasswd;
    }

    /**
     *  Set the type of the client keystore file.
     *
     *  @param    keyStoreType    Type of of the keystore file.
     */
    public void setKeyStoreType(String keyStoreType)
    {
        _keyStoreType = keyStoreType;
    }

    /**
     *  Set the truststore file required for server authentication.
     *
     *  @param    trustStoreFile    Location of the truststore file.
     */
    public void setTrustStore(String trustStoreFile)
    {
        _trustStoreFile = trustStoreFile;
    }

    /**
     *  Gets the client keystore file.
     */
    public String getKeyStoreFile()
    {
        return _keyStoreFile;
    }

    /**
     *  Gets the client keystore password.
     */
    public String getKeyStorePasswd()
    {
        return _keyStorePasswd;
    }

    /**
     *  Gets the client private key password.
     */
    public String getKeyPasswd()
    {
        return (_keyPasswd == null ? _keyStorePasswd : _keyPasswd);
    }

    /**
     *  Gets the type of the client keystore file.
     */
    public String getKeyStoreType()
    {
        return _keyStoreType;
    }

    /**
     *  Gets the location of the server truststore file.
     */
    public String getTrustStoreFile()
    {
        return _trustStoreFile;
    }

    /**
     *  Gets the type of the server truststore file.
     */
    public String getTrustStoreType()
    {
        return _trustStoreType;
    }

    /**
     *  Sets the hostname that is required for server authentication.
     * @param    host    Name of the host the client is connecting to.
     */
    public void setHost(String host)
    {
        _hostName  = host;
    }

    /**
     *  Gets the hostname that is required for server authentication.
     */
    public String getHost()
    {
        return _hostName;
    }

    /**
     *  Enables Certificate Based Authentication mechanism.
    */
    public void enableCBA()
    {
        _isCBAEnabled = true;
    }

    /**
     *  Disables Certificate Based Authentication mechanism.
    */
    public void disableCBA()
    {
        _isCBAEnabled = false;
        _isServerCertVerificationEnabled = false;
        _isHostnameVerificationEnabled = false;
    }

    /**
     *  Flag indicates whether to use Certificate Based Authentication mechanism.
    */
    public boolean isCBAEnabled()
    {
        return _isCBAEnabled;
    }

    /**
     *  Enables Server certificate verification by the client.
    */
    public void enableServerCertVerification()
    {
        _isServerCertVerificationEnabled = true;
        _isHostnameVerificationEnabled = true;
    }

    /**
     *  Disables Server certificate verification by the client.
    */
    public void disableServerCertVerification()
    {
        _isServerCertVerificationEnabled = false;
        _isHostnameVerificationEnabled = false;
    }

    /**
     *  Flag that indicates whether to use server certificate verification or not.
    */
    public boolean isServerCertVerificationEnabled()
    {
        return _isServerCertVerificationEnabled;
    }

    /**
     *  Enables Hostname verification by the client during server 
     *  certificate verification.
     */
    public void enableHostnameVerification()
    {
        _isHostnameVerificationEnabled = true;
    }

    /**
     *  Disables Hostname verification by the client during server 
     *  certificate verification.
     */
    public void disableHostnameVerification()
    {
        _isHostnameVerificationEnabled = false;
    }

    /**
     *  Determines whether Hostname verification is enabled by the 
     *  client during server certificate verification.
     */
    public boolean isHostnameVerificationEnabled()
    {
        return _isHostnameVerificationEnabled;
    }
    
    /**
     *  Compares the given authentication info objects according to their 
     * keystore and truststore properties. Returns true of both the properties matches.
    */
    public static boolean compare(AuthInfo a1, AuthInfo a2)
    {
        if (a1 == null && a2 == null) {
            return true;
        }
        else if (a1 == null && a2 != null) {
            return false;
        }
        else if (a1 != null && a2 == null) {
            return false;
        }
        else if ((a1.isServerCertVerificationEnabled() == a2.isServerCertVerificationEnabled())
                && (a1.isCBAEnabled() == a2.isCBAEnabled())
                && (a1.getKeyStoreFile() == a2.getKeyStoreFile())
                && (a1.getKeyStorePasswd() == a2.getKeyStorePasswd())
                && (a1.getKeyPasswd() == a2.getKeyPasswd())
                && (a1.getKeyStoreType() == a2.getKeyStoreType())
                && (a1.getTrustStoreFile() == a2.getTrustStoreFile())
                && (a1.getTrustStoreType() == a2.getTrustStoreType())) {
                return true;
        }
        else {
            return false;
        }
    }
}