/*
 * $Id: //depot/prod/zephyr/lufthansaair/src/java/netapp/manage/http/HTTPMessage.java#1 $
 *
 * Copyright (c) 2001-2004 Network Appliance, Inc.
 * All rights reserved.
 *
 * ONTAPI Server stuff
 */

package netapp.manage.http;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.io.EOFException;
import java.io.ByteArrayOutputStream;
import java.io.UnsupportedEncodingException;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import netapp.manage.NaCertificateException;
import netapp.manage.NaSSLException;

/**
 *  This class represents a HTTP message.
 */
public abstract class HTTPMessage {

	/**
	 *  A constant that specifies version HTTP/1.0.
	 */
	public static final int VERSION_1_0 = 1;

	/**
	 *  A constant that specifies version HTTP/1.1.
	 */
	public static final int VERSION_1_1 = 2;

	/**
	 *  A constant that specifies the HTTP GET method.
	 */
	public static final int METHOD_GET = 1;

	/**
	 *  A constant that specifies the HTTP POST method.
	 */
	public static final int METHOD_POST = 2;

	private static final byte[] CRLF = { (byte) '\r', (byte) '\n' };

	protected int _version;
	protected Map _headerLowerCase;
	protected Map _header;
	protected byte[] _contents;

	/**
	 *  Constructs a <code>HTTPMessage</code>.
	 *
	 *  @param	version		the HTTP version.
	 */
	protected HTTPMessage(int version)
	{
		_version = version;
		_headerLowerCase = new HashMap();
		_header = new HashMap();
		_contents = null;
	}

	/**
	 *  Set the contents of the message-body of this message. The
	 *  contents are represented as a byte array.
	 *
	 *  @param	contents	the message-body.
	 */
	public void setContents(byte[] contents)
	{
		_contents = contents;
	}

	/**
	 *  Return the contents of the message body of this message. The
	 *  contents are represented as a byte array.
	 *
	 *  @return		the message-body.
	 */
	public byte[] getContents()
	{
		return _contents;
	}

	/**
	 *  Fix the case of a header key. If this message contains a header
	 *  with a key that differs only in case from the key provided,
	 *  return the header key from this message. Otherwise, remember the
	 *  case of the provided key and return it.
	 *
	 *  @param	key	the header key to check.
	 *  @return		the correct header key.
	 */
	private String fixHeaderKey(String key)
	{
		String fixed = (String) _headerLowerCase.get(key.toLowerCase());
		if (null == fixed) {
			_headerLowerCase.put(key.toLowerCase(), key);
			fixed = key;
		}
		return fixed;
	}

	/**
	 *  Return the contents of one HTTP header from this message.
	 *
	 *  @param	key	which HTTP header.
	 *  @return		the value of the header.
	 */
	public String getHeader(String key)
	{
		key = fixHeaderKey(key);
		return (String) _header.get(key);
	}

	/**
	 *  Set the contents of one HTTP header in this message.
	 *
	 *  @param	key	which HTTP header.
	 *  @param	value	the new value of the header.
	 */
	public void setHeader(String key, String value)
	{
		key = fixHeaderKey(key);
		if(null == value) {
			_header.remove(key);
		} else {
			_header.put(key,value);
		}
	}

	/**
	 *  Remove one HTTP header from this message.
	 *
	 *  @param	key	which HTTP header.
	 */
	public String removeHeader(String key)
	{
		key = fixHeaderKey(key);
		return (String) _header.remove(key);
	}

	/**
	 *  Parse a HTTP version string.
	 *
	 *  @param	s	a version string.
	 *  @return		either {@link #VERSION_1_0} or
	 *			{@link #VERSION_1_1}.
	 *  @exception	HTTPException	if the string is not a valid HTTP
	 *				version string.
	 */
	protected static int versionFromString(String s) throws HTTPException
	{
		if (s.equalsIgnoreCase("HTTP/1.0")) {
			return VERSION_1_0;
		} else if (s.equalsIgnoreCase("HTTP/1.1")) {
			return VERSION_1_1;
		} else {
			throw new HTTPException("Invalid version: " + s);
		}
	}

	/**
	 *  Get the string representation of a HTTP version. Invalid
	 *  version are represented by the empty string.
	 *
	 *  @param	version	the version.
	 *  @return		string representation of the version.
	 */
	protected static String versionToString(int version)
	{
		switch (version) {
		case VERSION_1_0:
			return "HTTP/1.0";
		case VERSION_1_1:
			return "HTTP/1.1";
		default:
			return ""; /* HTTP/0.9 */
		}
	}

	/**
	 *  Parse a HTTP method string
	 *
	 *  @param	s	a method string.
	 *  @return		either {@link #METHOD_GET} or
	 *			{@link #METHOD_POST}.
	 *  @exception	HTTPException	if the string is not a valid HTTP
	 *				method string.
	 */
	protected static int methodFromString(String s) throws HTTPException
	{
		if (s.equalsIgnoreCase("GET")) {
			return METHOD_GET;
		} else if (s.equalsIgnoreCase("POST")) {
			return METHOD_POST;
		} else {
			throw new HTTPException("Unknown method: " + s);
		}
	}

	/**
	 *  Get the string representation of a HTTP method.
	 *
	 *  @param	method	the method.
	 *  @return		string representation of the method.
	 */
	protected static String methodToString(int method)
	{
		switch (method) {
		case METHOD_GET:
			return "GET";
		case METHOD_POST:
			return "POST";
		default:
			return "UNKNOWN";
		}
	}

	/**
	 *  Read one line of text. This is a utility function that reads
	 *  RFC2616 compliant lines of text from socket (using the platform's
	 *  default charset).
	 *
	 *  @param	in	the input stream (which is presumably connected
	 *			to a HTTP socket).
	 *  @return		a line of text.
	 *  @exception	IOException	if there is a communication problem on
	 *				the socket or if an invalid line terminator
	 *				is encountered.
	 */
	protected static String readLine(InputStream in) throws IOException
	{
		byte[] buffer = new byte[128];
		int nb = 0;
		String line;

		while (true) {
			int c = in.read();
			if (0 > c) {
				if (0 == nb) {
					throw new EOFException();
				}
				break;
			} else if('\n' == c)  {
				break;
			} else if('\r' == c)  {
				c = in.read();
				if ('\n' != c) {
					throw new IOException(
						"Invalid line terminator");
				}
				break;
			}
			buffer[nb++] = (byte) c;
			if (buffer.length < nb * 2) {
				byte[] b2 = new byte[buffer.length * 2];
				System.arraycopy(buffer, 0, b2, 0, nb);
				buffer = b2;
			}
		}

		line = (new String(buffer, 0, nb)).trim();

		if (0 < HTTPClient.snoop) {
			System.err.println("Read Line === " + line);
		}

		return line;
	}

	/**
	 *  Read the message-body of this message. This is a utility function
	 *  that reads the message-body from a socket assuming that the message
	 *  was created from the headers immediately preceding the message-body.
	 *  The message-body is stored in the contents of this message.
	 *
	 *  @param	in	the input stream (which is presumably connected
	 *			to a HTTP socket).
	 *  @exception	IOException	if there is a communication problem on
	 *				the socket.
	 */
	protected void readContents(InputStream in) throws IOException
	{
		String contentLength = getHeader("Content-Length");
		if (null != contentLength) {
			int length = Integer.parseInt(contentLength);
			byte[] buffer = new byte[length];
			int nb = 0;
			while (nb < length) {
				int n = in.read(buffer, nb, length - nb);
				if (0 > n) {
					throw new EOFException("Encountered EOF at " +
						nb + "/" + length + " bytes");
				}
				nb += n;
			}
			setContents(buffer);
		} else {
			ByteArrayOutputStream temp = new ByteArrayOutputStream();
			byte[] buffer = new byte[1024];
			while (true) {
				int n = in.read(buffer);
				if (0 > n) {
					break;
				}
				temp.write(buffer, 0, n);
			}
			setContents(temp.toByteArray());
		}
	}

	/**
	 *  Write this message. Write this entire HTTP message to a socket.
	 *
	 *  @param	out	the output stream stream (which is presumably
	 *			connected to a HTTP socket).
	 *  @exception	IOException	if there is a communication problem on
	 *				the socket.
	 */
	public void write(OutputStream out) throws IOException
	{
		byte[] headerBytes = toString().getBytes();
		int len = headerBytes.length + CRLF.length;

		if (null != _contents) {
			len += _contents.length;
		}

		byte[] buffer = new byte[len];

		System.arraycopy(headerBytes, 0, buffer, 0, headerBytes.length);
		System.arraycopy(CRLF, 0, buffer, headerBytes.length, CRLF.length);
		if (null != _contents) {
			System.arraycopy(_contents, 0, buffer, 
				headerBytes.length + CRLF.length, _contents.length);
		}
		try {
			out.write(buffer);
		} catch (IOException e) {
			if (e.getCause() instanceof NaCertificateException) {
				NaCertificateException certException = (NaCertificateException) e.getCause();
				throw new NaSSLException(certException.getCertificateChain(), certException.getMessage());
			}
		}
		out.flush();
		if (0 < HTTPClient.snoop) {
			String s;
			try {
				s = new String(_contents, "UTF-8");
			} catch (UnsupportedEncodingException e) {
				s = new String(_contents);
			}
			System.err.println("SENDING ======\n" + toString() + "\n" + s);
		}
	}

	/**
	 *  Set a header based on a line of text. Parses a line of text
	 *  representing a HTTP header, and store the header in this message.
	 *
	 *  @param	line		a line of text.
	 *  @exception	HTTPException	if the line is not a valid HTTP header.
	 */
	protected void processHeaderLine(String line) throws HTTPException
	{
		int split = line.indexOf(':');
		if (split < 0) {
			throw new HTTPException("Invalid header line: " +
				line);
		}
		_header.put(line.substring(0, split),
			line.substring(split + 1).trim());
	}

	/**
	 *  Get the first (status) line of this HTTP message.
	 *
	 *  @return     a string representing one line.
	 */
	protected abstract String firstLine();

	/**
	 *  Get the HTTP headers. Returns a string representing several lines
	 *  of text containing all the HTTP headers in this message.
	 *
	 *  @return     a string representing many lines of text.
	 */
	protected String headersString()
	{
		StringBuffer s = new StringBuffer();
		for (Iterator i = _header.keySet().iterator(); i.hasNext();) {
			Object key = i.next();
			s.append(key + ": " + _header.get(key) + "\r\n");
		}
		return s.toString();
	}

	/**
	 *  Return a <code>String</code> representation of this
	 *  <code>HTTPMessage</code>. The first line (status line) plus the
	 *  HTTP headers are included in the string.
	 *
	 *  @return     a string representing the first line and headers.
	 */
	public String toString()
	{
		return firstLine() + headersString();
	}
}
