/*
 * $Id: //depot/prod/zephyr/lufthansaair/src/java/netapp/manage/http/HTTPResponse.java#1 $
 *
 * Copyright (c) 2001-2004 Network Appliance, Inc.
 * All rights reserved.
 *
 * ONTAPI Server stuff
 */

package netapp.manage.http;

import java.io.InputStream;
import java.io.IOException;
import java.util.StringTokenizer;

/**
 *  This class represents a HTTP response.
 */
public class HTTPResponse extends HTTPMessage {

	private int _statusCode;
	private String _reasonPhrase;

	/**
	 *  Constructs a <code>HTTPResponse</code>.
	 *
	 *  @param	version		the HTTP version.
	 *  @param	statusCode	the HTTP status of the response.
	 *  @param	reasonPhrase	a textual description of the status (may be empty).
	 */
	public HTTPResponse(int version, int statusCode, String reasonPhrase)
	{
		super(version);
		_statusCode = statusCode;
		_reasonPhrase = reasonPhrase;
	}

	/**
	 *  Get the HTTP status of this response. It should be a 3 digit
	 *  decimal number.
	 *
	 *  @return     the HTTP status.
	 */
	public int getStatusCode()
	{
		return _statusCode;
	}

	/**
	 *  Get the textual status of this response. This is usually meant
	 *  for human readability, as {@link #getStatusCode()} contains all
	 *  the relevant information.
	 *
	 *  @return     the reason phrase (may be empty).
	 */
	public String getReasonPhrase()
	{
		return _reasonPhrase;
	}

	/**
	 *  Create a new HTTP response by reading data from a socket.
	 *
	 *  @param	in	an <code>InputStream</code>, presumably
	 *			connected to a client socket.
	 *  @return		a new <code>HTTPResponse</code>.
	 *  @exception	IOException	if there is a communication error on
	 *				the <code>InputStream</code>.
	 *  @exception	HTTPException	if the HTTP response is malformed.
	 */
	public static HTTPResponse read(InputStream in)
		throws IOException, HTTPException
	{
		String arguments = readLine(in);

		StringTokenizer st = new StringTokenizer(arguments);
		if (st.countTokens() < 2) {
			throw new HTTPException("Malformed response: "+  arguments);
		}

		int version = versionFromString(st.nextToken());
		int statusCode = Integer.parseInt(st.nextToken());
		String reasonPhrase = arguments.substring((versionToString(version)
			+ " " + statusCode).length()).trim();

		HTTPResponse response = new HTTPResponse(version, statusCode, reasonPhrase);

		for (String line = readLine(in); 0 < line.length();
				line = readLine(in)) {
			response.processHeaderLine(line);
		}

		response.readContents(in);

		return response;
	}

	/**
	 *  Determine whether this response indicates that the socket can
	 *  be used for additional requests.
	 *
	 *  @return     <code>true</code> if the socket can be reused,
	 *		otherwise <code>false</code>.
	 */
	public boolean canKeepAlive()
	{
		String contentLength = getHeader("Content-Length");
		String transferEncoding = getHeader("Transfer-Encoding");
		if (null == contentLength && null == transferEncoding) {
			return false;
		}
		String connection = getHeader("Connection");
		if (VERSION_1_1 == _version && (null == connection ||
				! connection.equalsIgnoreCase("close"))) {
			return true;
		} else if (VERSION_1_0 == _version && (null != connection &&
				connection.equalsIgnoreCase("Keep-Alive"))) {
			return true;
		}
		return false;
	}

	/**
	 *  Get the first (status) line of this HTTP response.
	 *
	 *  @return     a string representing one line.
	 */
	protected String firstLine()
	{
		return versionToString(_version) + " " + (_statusCode +
			" " + _reasonPhrase).trim() + "\r\n";
	}
}
